import { Algebra, Comonad, Foldable, Monoid, Semigroup } from './interfaces.js'

/** @import { InferredMorphism, Morphism } from './types.js' */

class Empty extends Algebra(Semigroup, Foldable, Monoid, Comonad) {
  constructor() {
    super()
  }

  /**
   * @template T
   * @this {Empty}
   * @param {T} value 
   * @returns {List<T>}
   */
  prepend(value) {
    return new Element(value, () => this)
  }

  head() {
    return empty
  }

  tail() {
    return empty
  }

  /**
   * @template T
   * @param {List<Element<T>>} other 
   * @returns {List<Element<T>>}
   */
  concat(other) {
    return other
  }

  chain(f) {
    return this
  }

  map(f) {
    return this
  }

}

/**
 * @template T
 */
class Element extends Algebra(Semigroup, Foldable, Monoid, Comonad) {
  #head
  #tail

  /**
   * @param {T} head
   * @param {() => List<T>} tail
   */
  constructor(head, tail = List.empty) {
    super()
    this.#head = head
    this.#tail = tail
  }

  /**
   * @template T
   * @this {Empty}
   * @param {T} value 
   * @returns {List<T>}
   */
  prepend(value) {
    return new Element(value, () => this)
  }

  head() {
    return this.#head
  }

  tail() {
    return this.#tail()
  }

  /**
   * @param {List<T>} other 
   * @returns {List<T>}
   */
  concat(other) {
    if (this.isEmpty()) {
      return other
    } else {
      return new Element(
        this.head(),
        () => this.tail().concat(other)
      )
    }
  }

  /**
   * @template U
   * @param {Morphism<T, List<U>>} f 
   * @returns {List<U>}
   */
  chain(f) {
    return new Element(
      f(this.head()),
      this.tail().chain(f)
    )
  }

  /**
   * @template U
   * @param {(acc: U, value: T) => U} f
   * @param {U} init
   * @returns {U}
   */
  reduce(f, init) {
    const acc = f(init, this._value)
    return this.tail().reduce(f, acc)
  }

  extract() {
    return this.reduce(reduceArray, [])
  }

  count() {
    return this.tail().count() + 1
  }

  /** @returns {this is Empty} */
  isEmpty() {
    return false
  }
}

/**
 * @template T
 * @type {Element<T> | Empty}
 */
export class List {
  /**
   * @template  T
   * @param {T} value
   * @returns {List<T>}
   */
  static of(value) {
    return empty.prepend(value)
  }

  /**
   * @template T
   * @param {Iterable<T>} iterable
   * @returns {List<T>}
   */
  static from(iterable) {
    const iterator = Iterator.from(iterable)
    return List.fromIterator(iterator)
  }

  /**
   * @template T
   * @param {Iterator<T>} iterator
   * @returns {List<T>}
   */
  static fromIterator(iterator) {
    const next = iterator.next()

    if (next.done) {
      return List.empty()
    } else {
      return new Element(next.value, () => List.fromIterator(iterator))
    }
  }

  /**
   * @template T
   * @param {T} head 
   * @param {List<T>} tail 
   * @returns {List<T>}
   */
  static cons(head, tail) {
    return new Element(head, () => tail)
  }

  static empty() {
    return Empty
  }
}

const empty = new Empty()

export const list = value => List.of(value)

